package main

import (
	"container/heap"
	"flag"
	"io/ioutil"
	"time"

	"gopkg.in/yaml.v2"
)

type ProxyOptions struct {
	HistorySize  int64 `yaml:"HistorySize"`
	NetworkSpeed int64 `yaml:"NetworkSpeed"`
}

type Options struct {
	// average value size including allocation overhead
	ValueSize        int64          `yaml:"ValueSize"`
	HistoryCacheSize int64          `yaml:"HistoryCacheSize"`
	NVPS             int64          `yaml:"NVPS"`
	Proxies          []ProxyOptions `yaml:"Proxies"`
	StopPerc         int            `yaml:"StopPerc"`
	ThrottlePerc     int            `yaml:"ThrottlePerc"`
	RunPerc          int            `yaml:"RunPerc"`
}

type Event interface {
	scheduled() int64
	do()
	next() Event
}

type Queue []Event

func (q Queue) Len() int { return len(q) }

func (q Queue) Less(i, j int) bool {
	return q[i].scheduled() < q[j].scheduled()
}

func (q Queue) Swap(i, j int) {
	q[i], q[j] = q[j], q[i]
}

func (q *Queue) Push(x interface{}) {
	*q = append(*q, x.(Event))
}

func (q *Queue) Pop() interface{} {
	old := *q
	n := len(old)
	event := old[n-1]
	old[n-1] = nil
	*q = old[0 : n-1]
	return event
}

func (q Queue) Peek() Event {
	if len(q) == 0 {
		return nil
	}
	return q[0].(Event)
}

var now int64

func emulate() {
	cache.init(&options)

	events := make(Queue, 0)
	now = time.Now().Unix() - 60*60*24

	for i := range options.Proxies {
		heap.Push(&events, &ProxyUpload{
			nextcheck: now + 1,
			size:      0,
			proxy:     newProxy(i, &options.Proxies[i]),
		})
	}

	heap.Push(&events, &CacheSync{
		nextcheck: now + 1,
	})

	for len(events) > 1 || cache.used > 0 {
		cache.lastUsage = cache.usage()
		for len(events) > 0 && events.Peek().scheduled() <= now {
			event := heap.Pop(&events).(Event)
			event.do()
			if event = event.next(); event != nil {
				heap.Push(&events, event)
			}
		}
		now++
		//		time.Sleep(time.Second)
	}
}

var options Options
var export *Export

func main() {
	var configFile, outputFile string
	var exportMask ExportMask

	flag.StringVar(&configFile, "m", "", "Configuration file")
	flag.Var(&exportMask, "e", "Export option, allowed values: discovery,history")
	flag.StringVar(&outputFile, "o", "", "Export output file name")
	flag.Parse()

	b, err := ioutil.ReadFile(configFile)
	if err != nil {
		panic(err)
	}

	err = yaml.Unmarshal(b, &options)
	if err != nil {
		panic(err)
	}

	export = newExport(exportMask, outputFile)
	defer export.close()

	export.writeDiscovery()

	if (export.mask & EXPORT_HISTORY) != 0 {
		emulate()
	}
}
